local lib = LibStub("LibAuraInfo-1.0");
if(not lib) then error("combatlog.lua for LibAuraInfo needs library loaded."); return; end

local debugPrint = lib.debugPrint
local GetTime = GetTime
local pairs = pairs
local table_remove = table.remove
local table_insert = table.insert
local bit_band = bit.band
local COMBATLOG_OBJECT_TYPE_PLAYER = COMBATLOG_OBJECT_TYPE_PLAYER
local GetSpellInfo = GetSpellInfo


lib.GUIDAuras = lib.GUIDAuras or {}

--I'm having a table for each info to minimize table creation.
lib.GUIDAuraData_auraFound = {}
lib.GUIDAuraData_casterGUID = {}
lib.GUIDAuraData_duration = {}
lib.GUIDAuraData_startTime = {}
lib.GUIDAuraData_expirationTime = {}
lib.GUIDAuraData_isDebuff = {}
lib.GUIDAuraData_debuffType = {}
lib.GUIDAuraData_stackCount = {}

--[===[@debug@
--Always track for me, but only track for end users if a addon has registered a callback.
--~ lib.trackAuras = true
--~ lib.frame:RegisterEvent('COMBAT_LOG_EVENT_UNFILTERED')
--@end-debug@]===]


function lib.frame:COMBAT_LOG_EVENT_UNFILTERED(event, ...)
	--timestamp, eventType, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags
	local _, eventType  = ...-- ***

	if self[eventType] then
		self[eventType](self, eventType, ...)
	end
end

function lib:AddAuraToGUID(dstGUID, spellID, srcGUID, isDebuff)
	local key = dstGUID.."-"..spellID
	local duration = self:GetDuration(spellID, srcGUID, dstGUID)
	local debuffType = self:GetDebuffType(spellID)
	
--~ 	local drDuration = self:GetDRDuration(dstGUID, spellID, duration)
--~ 	if duration ~= drDuration then
--~ 		local spellName = GetSpellInfo(spellID)
--~ 		debugPrint("DR duration mismatch!", spellName, "full:"..duration, "DR:"..drDuration)
--~ 	end
	
	local currentTime = GetTime()
	
	self.GUIDAuras[dstGUID] = self.GUIDAuras[dstGUID] or {}
	table_insert(self.GUIDAuras[dstGUID], #self.GUIDAuras[dstGUID]+1, spellID)
	
	
	self.GUIDAuraData_auraFound[key] = true
	self.GUIDAuraData_casterGUID[key] = srcGUID
	self.GUIDAuraData_duration[key] = duration
	self.GUIDAuraData_expirationTime[key] = currentTime + duration
	self.GUIDAuraData_isDebuff[key] = isDebuff
	self.GUIDAuraData_debuffType[key] = debuffType

--~ 	debugPrint("AddAuraToGUID", dstGUID, spellID)
	
end



local function RemoveAuraFromGUID(dstGUID, spellID)
	for i= #lib.GUIDAuras[dstGUID], 1, -1 do 
		if lib.GUIDAuras[dstGUID][i] == spellID then
			table_remove(lib.GUIDAuras[dstGUID], i)
			break
		end
	end
	
	local key = dstGUID.."-"..spellID
	lib.GUIDAuraData_auraFound[key] = false
--~ 	lib.GUIDAuraData_casterGUID[key] = nil
--~ 	lib.GUIDAuraData_duration[key] = nil
--~ 	lib.GUIDAuraData_expirationTime[key] = nil
--~ 	lib.GUIDAuraData_isDebuff[key] = nil
--~ 	lib.GUIDAuraData_debuffType[key] = nil
--~ 	lib.GUIDAuraData_stackCount[key] = nil
end

function lib:RemoveAllAurasFromGUID(dstGUID)
	if self.GUIDAuras[dstGUID] then
		local spellID
		local key
		for i= #self.GUIDAuras[dstGUID], 1, -1 do 
			spellID = self.GUIDAuras[dstGUID][i]
			key = dstGUID.."-"..spellID
			
			self.GUIDAuraData_auraFound[key] = false
--~ 			self.GUIDAuraData_casterGUID[key] = nil
--~ 			self.GUIDAuraData_duration[key] = nil
--~ 			self.GUIDAuraData_expirationTime[key] = nil
--~ 			self.GUIDAuraData_isDebuff[key] = nil
--~ 			self.GUIDAuraData_debuffType[key] = nil
--~ 			self.GUIDAuraData_stackCount[key] = nil
			
			table_remove(self.GUIDAuras[dstGUID], i)
		end
	end
end

local function AuraAlreadyOnGUID(dstGUID, spellID, casterGUID)
	local key = dstGUID.."-"..spellID
	if lib.GUIDAuraData_casterGUID[key] and lib.GUIDAuraData_casterGUID[key] == casterGUID then
		return true
	end
	return false
end

lib.GUIDData_name = {}
lib.GUIDData_flags = {}
local function SaveGUIDInfo(guid, name, flags)
	lib.GUIDData_name[guid] = name
	lib.GUIDData_flags[guid] = flags
end

function lib.frame:SPELL_AURA_APPLIED(event, timestamp, eventType, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)	--
	local spellID, spellName, spellSchool, auraType, amount  = ...
	
	SaveGUIDInfo(srcGUID, srcName, srcFlags)
	SaveGUIDInfo(dstGUID, dstName, dstFlags)
	
	if lib.drSpells[spellID] then
		lib:GUIDGainedDRAura(dstGUID, spellID, lib:FlagIsPlayer(dstFlags))
	end
	
	if lib.auraInfo[spellID] then
--~ 		local duration, debuffType = self:GetSpellInfo(spellID)
		lib:RemoveOldAuras(dstGUID)

--~ 		lib.GUIDAuras[dstGUID] = lib.GUIDAuras[dstGUID] or {}
--~ 		lib.GUIDAuraData[dstGUID] = lib.GUIDAuraData[dstGUID] or {}
		
--~ 		debugPrint(event, dstName, spellName)
		
		if not lib.GUIDAuras[dstGUID] or #lib.GUIDAuras[dstGUID] == 0 then
			lib:AddAuraToGUID(dstGUID, spellID, srcGUID, auraType == "DEBUFF")
			lib.callbacks:Fire("LibAuraInfo_AURA_APPLIED", dstGUID, spellID, srcGUID, spellSchool, auraType)
			return
		end
		
		if not AuraAlreadyOnGUID(dstGUID, spellID, srcGUID) then
			lib:AddAuraToGUID(dstGUID, spellID, srcGUID, auraType == "DEBUFF")
			lib.callbacks:Fire("LibAuraInfo_AURA_APPLIED", dstGUID, spellID, srcGUID, spellSchool, auraType)
		end
		
	end
	
end



function lib.frame:SPELL_AURA_REMOVED(event, timestamp, eventType, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)	--
	local spellID, spellName, spellSchool, auraType  = ...
	if lib.GUIDAuras[dstGUID] then
		local key = dstGUID.."-"..spellID
		if lib.GUIDAuraData_casterGUID[key] and lib.GUIDAuraData_casterGUID[key] == srcGUID then
			RemoveAuraFromGUID(dstGUID, spellID)
			lib.callbacks:Fire("LibAuraInfo_AURA_REMOVED", dstGUID, spellID, srcGUID, spellSchool, auraType)
		end
	end
	
	if lib.drSpells[spellID] then
		lib:GUIDRemovedDRAura(dstGUID, spellID, lib:FlagIsPlayer(dstFlags))
	end
end

function lib.frame:SPELL_AURA_REFRESH(event, timestamp, eventType, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)
	local spellID, spellName, spellSchool, auraType  = ...
	
	if lib.drSpells[spellID] then
		lib:GUIDRemovedDRAura(dstGUID, spellID, lib:FlagIsPlayer(dstFlags))
		lib:GUIDGainedDRAura(dstGUID, spellID, lib:FlagIsPlayer(dstFlags))
	end
	
	if lib.GUIDAuras[dstGUID] then
		local key = dstGUID.."-"..spellID
--~ 		local duration = lib:GetDuration(spellID, srcGUID)
		local duration = lib:GetDuration(spellID, srcGUID, dstGUID)
--~ 		if lib.GUIDAuraData_casterGUID[key] and lib.GUIDAuraData_casterGUID[key] == srcGUID then
--~ 			lib.GUIDAuraData_expirationTime[key] = GetTime() + duration
			
			lib.GUIDAuraData_expirationTime[key] = GetTime() + duration
			lib.GUIDAuraData_casterGUID[key] = srcGUID
			
			
			lib.callbacks:Fire("LibAuraInfo_AURA_REFRESH", dstGUID, spellID, srcGUID, spellSchool, auraType, lib.GUIDAuraData_expirationTime[key])
			return
--~ 		end
	end

--~ 	debugPrint("SPELL_AURA_REFRESH", dstName, spellName, "passing to SPELL_AURA_APPLIED")
	self:SPELL_AURA_APPLIED(event, timestamp, eventType, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)
end

--DOSE = spell stacking
function lib.frame:SPELL_AURA_APPLIED_DOSE(event, timestamp, eventType, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)
	local spellID, spellName, spellSchool, auraType  = ...

--~ 	if spellInfo[spellID] and guidBuffs[dstGUID] then
--~ 		for i=table_getn(guidBuffs[dstGUID]), 1, -1 do 
--~ 			if guidBuffs[dstGUID][i].sID == spellID and (not guidBuffs[dstGUID][i].caster or guidBuffs[dstGUID][i].caster == srcName) then
--~ 				guidBuffs[dstGUID][i].stackCount = guidBuffs[dstGUID][i].stackCount + 1
--~ 				
--~ 				if not self:UpdatePlateByGUID(dstGUID) and self:FlagIsPlayer(dstFlags) then
--~ 					self:UpdatePlateByName(dstName)
--~ 				end
--~ 				return
--~ 			end
--~ 		end
--~ 	end
	
	if lib.GUIDAuras[dstGUID] then
		local key = dstGUID.."-"..spellID
		if lib.GUIDAuraData_casterGUID[key] and lib.GUIDAuraData_casterGUID[key] == srcGUID then
			lib.GUIDAuraData_stackCount[key] = (lib.GUIDAuraData_stackCount[key] or 1) + 1
		
			lib.callbacks:Fire("LibAuraInfo_AURA_APPLIED_DOSE", dstGUID, spellID, srcGUID, spellSchool, auraType, lib.GUIDAuraData_stackCount[key])
			return
		end
	end

--~ 	debugPrint("SPELL_AURA_APPLIED_DOSE", dstName, spellName, "passing to SPELL_AURA_APPLIED")
	--Spell isn't in our list, let's add it.
	--Note this event could have fired on the 5th stack but our spell frame will only show it applied once. 
	self:SPELL_AURA_APPLIED(event, timestamp, eventType, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)
end

function lib.frame:SPELL_AURA_BROKEN_SPELL(...)
--~ 	local spellID, spellName, spellSchool, auraType  = ...
	self:SPELL_AURA_REMOVED(...)
end

function lib.frame:SPELL_AURA_BROKEN(...)
--~ 	local spellID, spellName, spellSchool, auraType  = ...
	self:SPELL_AURA_REMOVED(...)
end


function lib.frame:UNIT_DIED(event, timestamp, eventType, srcGUID, srcName, srcFlags, dstGUID, dstName, dstFlags, ...)
	if lib.GUIDAuras[dstGUID] then
		lib:RemoveAllAurasFromGUID(dstGUID)
		lib.callbacks:Fire("LibAuraInfo_AURA_CLEAR", dstGUID)
	end
end

function lib.frame:UNIT_DESTROYED(...)
	self:UNIT_DIED(...)
end

function lib.frame:UNIT_DISSIPATES(...)
	self:UNIT_DIED(...)
end

function lib.frame:PARTY_KILL(...)
	self:UNIT_DIED(...)
end

function lib:RemoveOldAuras(dstGUID)
	if lib.GUIDAuras[dstGUID] then
		local spellID
		local key
		local currentTime = GetTime()
		for i= #lib.GUIDAuras[dstGUID], 1, -1 do 
			spellID = lib.GUIDAuras[dstGUID][i]
			key = dstGUID.."-"..spellID
			
			if self.GUIDAuraData_expirationTime[key] and self.GUIDAuraData_expirationTime[key] > 0 and currentTime > self.GUIDAuraData_expirationTime[key] then
				RemoveAuraFromGUID(dstGUID, spellID)
			end
		end
	end
end

function lib:FlagIsPlayer(flags)
	if bit_band(flags, COMBATLOG_OBJECT_TYPE_PLAYER) == COMBATLOG_OBJECT_TYPE_PLAYER then
		return true
	end
	return nil
end
